import java.util.function.BiFunction;
import java.util.function.Function;

/**
 * La classe MCD implementa la soluzione dell'esercizio, ovvero
 * il calcolo del massimo comun divisore utilizzando mediante
 * l'algoritmo di Euclide. (Tale implementazione sfrutta la
 * metodologia per lo sviluppo di Function<T,U> ricorsive.)
 */
public class MCD {

    // Classe di supporto utilizzata per rappresentare una coppia
    // di numeri interi <n,m>.
    private static class CoppiaInteri {
        private int n;
        private int m;

        public CoppiaInteri(int n, int m) {
            this.n = n;
            this.m = m;
        }

        public int getN() { return this.n; }

        public int getM() { return this.m; }
    }

    // Funzione che implmenta l'algoritmo di Euclide utilizzando
    // la metodologia per lo sviluppo di Function<T,U> ricorsive
    private static Function<CoppiaInteri,Integer> mcd = c -> {
        // La funzione helper (annidata in mcd) implementa
        // l'algoritmo di Euclide
        BiFunction<BiFunction, CoppiaInteri, Integer> helper = (f, x) -> {
            // MCD(n,m) = n, se m=0
            if (x.getM() == 0) return x.getN();
            // MCD(n,m) = MCD(m,n), se m > n;
            if (x.getM() > x.getN())
                return (int) f.apply(f, new CoppiaInteri(x.getM(), x.getN()));
            // MCD(n,m) = MCD(n-m,m), altrimenti
            return (int) f.apply(f, new CoppiaInteri(x.getN() - x.getM(), x.getM()));
        };
        // La funzione mcd invoca opportunamente helper per calcolare
        // il massimo comun divisore tra n e m
        return helper.apply(helper, c);
    };

    // Metodo statico che consente il calcolo del massimo comun divisore
    // tra n e m (il cui calcolo è avviene invocando la funzione mcd
    // dichiarata sopra)
    public static int calcola(int n, int m) {
        return mcd.apply(new CoppiaInteri(n,m));
    }

    // Metodo statico (eseguibile) per testare il metodo calcola
    // (e quindi la funzione mcd implementata)
    public static void main(String[] args) {
        // Per configurare il test, è sufficiente modificare
        // i valori assegnati a n e m
        int n = 24;
        int m = 104;
        // Calcolo e visualizzazione di MCD(n,m)
        System.out.println(MCD.calcola(n,m));
    }
}
